from __future__ import absolute_import, unicode_literals

import datetime
import logging
import pytz

from antlr4 import InputStream, CommonTokenStream, ParseTreeVisitor, Token
from antlr4.error.Errors import ParseCancellationException, NoViableAltException
from antlr4.error.ErrorStrategy import BailErrorStrategy
from decimal import Decimal
from enum import Enum
from . import conversions, EvaluationError
from .dates import DateStyle, DateParser
from .functions import FunctionManager, custom, excel
from .utils import decimal_pow, urlquote, parse_json_date

logger = logging.getLogger(__name__)


DEFAULT_FUNCTION_MANAGER = FunctionManager()
DEFAULT_FUNCTION_MANAGER.add_library(excel)
DEFAULT_FUNCTION_MANAGER.add_library(custom)


class EvaluationContext(object):
    """
    Evaluation context, i.e. variables and date options
    """
    def __init__(self, variables=None, timezone=pytz.UTC, date_style=DateStyle.DAY_FIRST, now=None):
        self.variables = variables if variables is not None else {}
        self.timezone = timezone
        self.date_style = date_style
        self.now = now if now else datetime.datetime.now(timezone)

    @classmethod
    def from_json(cls, json_obj):
        variables = json_obj['variables']
        timezone = pytz.timezone(json_obj['timezone'])
        date_style = DateStyle.DAY_FIRST if json_obj['date_style'] == 'day_first' else DateStyle.MONTH_FIRST
        now = parse_json_date(json_obj.get('now'))

        return EvaluationContext(variables, timezone, date_style, now)

    def resolve_variable(self, path):
        return self._resolve_variable_in_container(self.variables, path.lower(), path)

    def put_variable(self, key, value):
        self.variables[key] = value

    def get_date_format(self, inc_time):
        if self.date_style == DateStyle.DAY_FIRST:
            pattern = "%d-%m-%Y"
        else:
            pattern = "%m-%d-%Y"

        if inc_time:
            pattern += " %H:%M"

        return pattern

    def get_date_parser(self):
        return DateParser(datetime.date.today(), self.timezone, self.date_style)

    def _resolve_variable_in_container(self, container, path, original_path):
        if '.' in path:
            (item, remaining_path) = path.split('.', 1)
        else:
            item = path
            remaining_path = None

        # copy of container with all lowercase keys
        container = {k.lower(): v for k, v in container.iteritems()}

        if item not in container:
            raise EvaluationError("Undefined variable: %s" % original_path)

        value = container[item]

        if remaining_path is not None and value is not None:
            if not isinstance(value, dict):
                raise EvaluationError("Undefined variable: %s" % original_path)

            return self._resolve_variable_in_container(value, remaining_path, original_path)

        elif isinstance(value, dict):
            if '*' in value:
                return value['*']
            elif '__default__' in value:
                return value['__default__']
            else:
                raise EvaluationError("Undefined variable: %s" % original_path)
        else:
            return value


class EvaluationStrategy(Enum):
    COMPLETE = 1           # evaluate all expressions completely
    RESOLVE_AVAILABLE = 2  # if expression contains missing context references, just substitute what variables we have


class State(Enum):
    """
    Temba templates support 2 forms of embedded expression:
      1. Single variable, e.g. @contact, @contact.name (delimited by character type or end of input)
      2. Contained expression, e.g. @(SUM(1, 2) + 2) (delimited by balanced parentheses)
    """
    BODY = 0            # not in a expression
    PREFIX = 1          # '=' prefix that denotes the start of an expression
    IDENTIFIER = 2      # the identifier part, e.g. 'contact.age' in '@contact.age'
    BALANCED = 3        # the balanced parentheses delimited part, e.g. '(1 + 2)' in '@(1 + 2)'
    STRING_LITERAL = 4  # a string literal
    ESCAPED_PREFIX = 5  # a '@' prefix preceded by another '@'


class Evaluator(object):

    def __init__(self, expression_prefix='@', allowed_top_levels=(), function_manager=DEFAULT_FUNCTION_MANAGER):
        """
        Creates a new evaluator
        :param expression_prefix: the prefix for expressions, e.g. @
        :param allowed_top_levels: top-level context items allowed outside of parentheses, e.g. contact, flow
        :param function_manager: the function manager to use
        """
        self._expression_prefix = expression_prefix
        self._allowed_top_levels = set(allowed_top_levels)
        self._function_manager = function_manager

    def evaluate_template(self, template, context, url_encode=False, strategy=EvaluationStrategy.COMPLETE):
        """
        Evaluates a template string, e.g. "Hello @contact.name you have @(contact.reports * 2) reports"
        :param template: the template string
        :param context: the evaluation context
        :param url_encode: whether or not values should be URL encoded
        :return: a tuple of the evaluated template and a list of evaluation errors
        """
        input_chars = list(template)
        output_chars = []
        errors = []
        state = State.BODY
        current_expression_chars = []
        current_expression_terminated = False
        parentheses_level = 0

        # determines whether the given character is a word character, i.e. \w in a regex
        is_word_char = lambda c: c and (c.isalnum() or c == '_')

        for pos, ch in enumerate(input_chars):
            # in order to determine if the b in a.b terminates an identifier, we have to peek two characters ahead as it
            # could be a.b. (b terminates) or a.b.c (b doesn't terminate)
            next_ch = input_chars[pos + 1] if (pos < (len(input_chars) - 1)) else None
            next_next_ch = input_chars[pos + 2] if (pos < (len(input_chars) - 2)) else None

            if state == State.BODY:
                if ch == self._expression_prefix and (is_word_char(next_ch) or next_ch == '('):
                    state = State.PREFIX
                    current_expression_chars = [ch]
                elif ch == self._expression_prefix and next_ch == self._expression_prefix:
                    state = State.ESCAPED_PREFIX
                else:
                    output_chars.append(ch)

            elif state == State.PREFIX:
                if is_word_char(ch):
                    # we're parsing an expression like @XXX
                    state = State.IDENTIFIER
                elif ch == '(':
                    # we're parsing an expression like @(1 + 2)
                    state = State.BALANCED
                    parentheses_level += 1

                current_expression_chars.append(ch)

            elif state == State.IDENTIFIER:
                current_expression_chars.append(ch)

            elif state == State.BALANCED:
                if ch == '(':
                    parentheses_level += 1
                elif ch == ')':
                    parentheses_level -= 1
                elif ch == '"':
                    state = State.STRING_LITERAL

                current_expression_chars.append(ch)

                # expression terminates if parentheses balance
                if parentheses_level == 0:
                    current_expression_terminated = True

            elif state == State.STRING_LITERAL:
                if ch == '"':
                    state = State.BALANCED
                current_expression_chars.append(ch)

            elif state == State.ESCAPED_PREFIX:
                state = State.BODY
                output_chars.append(ch)

            # identifier can terminate expression in 3 ways:
            #  1. next char is null (i.e. end of the input)
            #  2. next char is not a word character or period
            #  3. next char is a period, but it's not followed by a word character
            if state == State.IDENTIFIER:
                if not next_ch or (not is_word_char(next_ch) and next_ch != '.') or (next_ch == '.' and not is_word_char(next_next_ch)):
                    current_expression_terminated = True

            if current_expression_terminated:
                expression = ''.join(current_expression_chars)
                output_chars.append(self._resolve_expression_block(expression, context, url_encode, strategy, errors))
                current_expression_chars = []
                current_expression_terminated = False
                state = State.BODY

        # if last expression didn't terminate - add to output as is
        if not current_expression_terminated and current_expression_chars:
            output_chars.append(''.join(current_expression_chars))

        output = ''.join(output_chars)  # joining is fastest way to build strings in Python
        return output, errors

    def _resolve_expression_block(self, expression, context, url_encode, strategy, errors):
        """
        Resolves an expression block found in the template, e.g. @(...). If an evaluation error occurs, expression is
        returned as is.
        """
        try:
            body = expression[1:]  # strip prefix

            # if expression doesn't start with ( then check it's an allowed top level context reference
            if not body.startswith('('):
                top_level = body.split('.')[0].lower()
                if top_level not in self._allowed_top_levels:
                    return expression

            evaluated = self.evaluate_expression(body, context, strategy)

            # convert result to string
            result = conversions.to_string(evaluated, context)

            return urlquote(result) if url_encode else result
        except EvaluationError, e:
            logger.debug("EvaluationError: %s" % e.message)

            # if we can't evaluate expression, include it as is in the output
            errors.append(e.message)
            return expression

    def evaluate_expression(self, expression, context, strategy=EvaluationStrategy.COMPLETE):
        """
        Evaluates a single expression, e.g. "contact.reports * 2"
        :param expression: the expression string
        :param context: the evaluation context
        :param strategy: the evaluation strategy
        :return: the evaluated expression value
        """
        from .gen.ExcellentLexer import ExcellentLexer
        from .gen.ExcellentParser import ExcellentParser

        stream = InputStream(expression)
        lexer = ExcellentLexer(stream)
        tokens = CommonTokenStream(lexer)

        parser = ExcellentParser(tokens)
        parser._errHandler = BailErrorStrategy()

        try:
            tree = parser.parse()

            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Expression '%s' parsed as %s" % (expression, tree.toStringTree()))
        except ParseCancellationException, ex:
            message = None
            if ex.args and isinstance(ex.args[0], NoViableAltException):
                token = ex.args[0].offendingToken
                if token is not None and token.type != ExcellentParser.EOF:
                    message = "Expression error at: %s" % token.text

            if message is None:
                message = "Expression is invalid"

            raise EvaluationError(message, ex)

        if strategy == EvaluationStrategy.RESOLVE_AVAILABLE:
            resolved = self._resolve_available(tokens, context)
            if resolved is not None:
                return resolved

        visitor = ExcellentVisitor(self._function_manager, context)
        return visitor.visit(tree)

    def _resolve_available(self, tokens, context):
        """
        Checks the token stream for context references and if there are missing references - substitutes available
        references and returns a partially evaluated expression.
        :param tokens: the token stream (all tokens fetched)
        :param context: the evaluation context
        :return: the partially evaluated expression or none if expression can be fully evaluated
        """
        from .gen.ExcellentParser import ExcellentParser

        has_missing = False
        output_components = []

        for t in range(len(tokens.tokens) - 1):  # we can ignore the final EOF token
            token = tokens.get(t)
            next_token = tokens.get(t + 1)

            # if token is a NAME not followed by ( then it's a context reference
            if token.type == ExcellentParser.NAME and next_token.type != ExcellentParser.LPAREN:
                try:
                    output_components.append(context.resolve_variable(token.text))
                except EvaluationError:
                    has_missing = True
                    output_components.append(token)
            else:
                output_components.append(token)

        # if we don't have missing context references, perform evaluation as normal
        if not has_missing:
            return None

        # re-combine the tokens and context values back into an expression
        output = [self._expression_prefix]

        for output_component in output_components:
            if isinstance(output_component, Token):
                comp_val = output_component.text
            else:
                comp_val = conversions.to_repr(output_component, context)
            output.append(comp_val)

        return ''.join(output)


class ExcellentVisitor(ParseTreeVisitor):

    def __init__(self, functions, eval_context):
        self._functions = functions
        self._eval_context = eval_context

    def visitParse(self, ctx):
        return self.visit(ctx.expression())

    def visitFunctionCall(self, ctx):
        """
        expression : fnname LPAREN parameters? RPAREN
        """
        func_name = ctx.fnname().getText()

        if ctx.parameters() is not None:
            parameters = self.visit(ctx.parameters())
        else:
            parameters = []

        return self._functions.invoke_function(self._eval_context, func_name, parameters)

    def visitFunctionParameters(self, ctx):
        """
        parameters : expression (COMMA expression)*
        """
        return [self.visit(expression) for expression in ctx.expression()]

    def visitNegation(self, ctx):
        """
        expression: MINUS expression
        """
        return -conversions.to_decimal(self.visit(ctx.expression()), self._eval_context)

    def visitExponentExpression(self, ctx):
        """
        expression: expression EXPONENT expression
        """
        arg1 = conversions.to_decimal(self.visit(ctx.expression(0)), self._eval_context)
        arg2 = conversions.to_decimal(self.visit(ctx.expression(1)), self._eval_context)
        return conversions.to_decimal(decimal_pow(arg1, arg2), ctx)

    def visitMultiplicationOrDivisionExpression(self, ctx):
        """
        expression: expression (TIMES | DIVIDE) expression
        """
        is_mul = ctx.TIMES() is not None

        arg1 = conversions.to_decimal(self.visit(ctx.expression(0)), self._eval_context)
        arg2 = conversions.to_decimal(self.visit(ctx.expression(1)), self._eval_context)

        if not is_mul and arg2 == Decimal(0):
            raise EvaluationError("Division by zero")

        return arg1 * arg2 if is_mul else arg1 / arg2

    def visitAdditionOrSubtractionExpression(self, ctx):
        """
        expression: expression (PLUS | MINUS) expression
        """
        is_add = ctx.PLUS() is not None
        arg1 = self.visit(ctx.expression(0))
        arg2 = self.visit(ctx.expression(1))

        # first try as decimals
        try:
            _arg1 = conversions.to_decimal(arg1, self._eval_context)
            _arg2 = conversions.to_decimal(arg2, self._eval_context)
            return _arg1 + _arg2 if is_add else _arg1 - _arg2
        except EvaluationError:
            pass

        # then as date + something
        try:
            _arg1 = conversions.to_date_or_datetime(arg1, self._eval_context)

            if isinstance(arg2, datetime.time):
                # upgrade our date to datetime
                _arg1 = conversions.to_datetime(_arg1, self._eval_context)

                # convert time value to a duration
                _arg2 = datetime.timedelta(hours=arg2.hour, minutes=arg2.minute, seconds=arg2.second, microseconds=arg2.microsecond)
            else:
                _arg2 = datetime.timedelta(days=conversions.to_integer(arg2, self._eval_context))

            return _arg1 + _arg2 if is_add else _arg1 - _arg2

        except EvaluationError, ex:
            raise EvaluationError("Expression could not be evaluated as decimal or date arithmetic", ex)

    def visitComparisonExpression(self, ctx):
        """
        expression: expression (LTE | LT | GTE | GT) expression
        """
        arg1, arg2 = conversions.to_same(self.visit(ctx.expression(0)), self.visit(ctx.expression(1)), self._eval_context)

        if isinstance(arg1, basestring):
            # string comparison is case-insensitive
            compared = cmp(arg1.lower(), arg2.lower())
        else:
            compared = cmp(arg1, arg2)

        if ctx.LTE() is not None:
            return compared <= 0
        elif ctx.LT() is not None:
            return compared < 0
        elif ctx.GTE() is not None:
            return compared >= 0
        else:  # GT
            return compared > 0

    def visitEqualityExpression(self, ctx):
        """
        expression: expression (EQ | NEQ) expression
        """
        arg1, arg2 = conversions.to_same(self.visit(ctx.expression(0)), self.visit(ctx.expression(1)), self._eval_context)

        if isinstance(arg1, basestring):
            # string equality is case-insensitive
            equal = arg1.lower() == arg2.lower()
        else:
            equal = arg1 == arg2

        return equal if ctx.EQ() is not None else not equal

    def visitConcatenation(self, ctx):
        """
        expression: expression AMPERSAND expression
        """
        arg1 = conversions.to_string(self.visit(ctx.expression(0)), self._eval_context)
        arg2 = conversions.to_string(self.visit(ctx.expression(1)), self._eval_context)
        return arg1 + arg2

    def visitStringLiteral(self, ctx):
        """
        expression: STRING
        """
        value = ctx.getText()[1:-1]  # strip surrounding quotes
        return value.replace('""', '"')  # unescape embedded quotes

    def visitDecimalLiteral(self, ctx):
        """
        expression: DECIMAL
        """
        return Decimal(ctx.getText())

    def visitTrue(self, ctx):
        """
        expression: TRUE
        """
        return True

    def visitFalse(self, ctx):
        """
        expression: FALSE
        """
        return False

    def visitContextReference(self, ctx):
        """
        expression: NAME
        """
        identifier = ctx.NAME().getText()
        value = self._eval_context.resolve_variable(identifier)
        return value if value is not None else ""  # return empty string rather than none

    def visitParentheses(self, ctx):
        """
        expression: LPAREN expression RPAREN
        """
        return self.visit(ctx.expression())



from __future__ import absolute_import, unicode_literals

import datetime
import math
import pytz
import regex
import urllib

from decimal import Decimal, ROUND_HALF_UP

JSON_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'


def decimal_pow(number, power):
    """
    Pow for two decimals
    """
    return Decimal(math.pow(number, power))


def decimal_round(number, num_digits, rounding=ROUND_HALF_UP):
    """
    Rounding for decimals with support for negative digits
    """
    exp = Decimal(10) ** -num_digits

    if num_digits >= 0:
        return number.quantize(exp, rounding)
    else:
        return exp * (number / exp).to_integral_value(rounding)


def urlquote(text):
    """
    Encodes text for inclusion in a URL query string. Should be equivalent to Django's urlquote function.
    :param text: the text to encode
    :return: the encoded text
    """
    return urllib.quote(text.encode('utf-8'))


def tokenize(text):
    """
    Tokenizes a string by splitting on non-word characters.
    """
    splits = regex.split(r"\W+", text, flags=regex.UNICODE | regex.V0)
    return [split for split in splits if split]   # return only non-empty


def parse_json_date(value):
    """
    Parses an ISO8601 formatted datetime from a string value
    """
    if not value:
        return None

    return datetime.datetime.strptime(value, JSON_DATETIME_FORMAT).replace(tzinfo=pytz.UTC)


def format_json_date(value):
    """
    Formats a datetime as ISO8601 in UTC with millisecond precision, e.g. "2014-10-03T09:41:12.790Z"
    """
    if not value:
        return None

    # %f will include 6 microsecond digits
    micro_precision = value.astimezone(pytz.UTC).strftime(JSON_DATETIME_FORMAT)

    # only keep the milliseconds portion of the second fraction
    return micro_precision[:-4] + 'Z'




from __future__ import absolute_import, unicode_literals

import datetime

from decimal import Decimal, getcontext, ROUND_HALF_UP
from . import EvaluationError


def to_boolean(value, ctx):
    """
    Tries conversion of any value to a boolean
    """
    if isinstance(value, bool):
        return value
    elif isinstance(value, int):
        return value != 0
    elif isinstance(value, Decimal):
        return value != Decimal(0)
    elif isinstance(value, basestring):
        value = value.lower()
        if value == 'true':
            return True
        elif value == 'false':
            return False
    elif isinstance(value, datetime.date) or isinstance(value, datetime.time):
        return True

    raise EvaluationError("Can't convert '%s' to a boolean" % unicode(value))


def to_integer(value, ctx):
    """
    Tries conversion of any value to an integer
    """
    if isinstance(value, bool):
        return 1 if value else 0
    elif isinstance(value, int):
        return value
    elif isinstance(value, Decimal):
        try:
            val = int(value.to_integral_exact(ROUND_HALF_UP))
            if isinstance(val, int):
                return val
        except ArithmeticError:
            pass
    elif isinstance(value, basestring):
        try:
            return int(value)
        except ValueError:
            pass

    raise EvaluationError("Can't convert '%s' to an integer" % unicode(value))


def to_decimal(value, ctx):
    """
    Tries conversion of any value to a decimal
    """
    if isinstance(value, bool):
        return Decimal(1) if value else Decimal(0)
    elif isinstance(value, int):
        return Decimal(value)
    elif isinstance(value, Decimal):
        return value
    elif isinstance(value, basestring):
        try:
            return Decimal(value)
        except Exception:
            pass

    raise EvaluationError("Can't convert '%s' to a decimal" % unicode(value))


def to_string(value, ctx):
    """
    Tries conversion of any value to a string
    """
    if isinstance(value, bool):
        return "TRUE" if value else "FALSE"
    elif isinstance(value, int):
        return unicode(value)
    elif isinstance(value, Decimal):
        return format_decimal(value)
    elif isinstance(value, basestring):
        return value
    elif type(value) == datetime.date:
        return value.strftime(ctx.get_date_format(False))
    elif isinstance(value, datetime.time):
        return value.strftime('%H:%M')
    elif isinstance(value, datetime.datetime):
        return value.astimezone(ctx.timezone).strftime(ctx.get_date_format(True))

    raise EvaluationError("Can't convert '%s' to a string" % unicode(value))


def to_date(value, ctx):
    """
    Tries conversion of any value to a date
    """
    if isinstance(value, basestring):
        temporal = ctx.get_date_parser().auto(value)
        if temporal is not None:
            return to_date(temporal, ctx)
    elif type(value) == datetime.date:
        return value
    elif isinstance(value, datetime.datetime):
        return value.date()  # discard time

    raise EvaluationError("Can't convert '%s' to a date" % unicode(value))


def to_datetime(value, ctx):
    """
    Tries conversion of any value to a datetime
    """
    if isinstance(value, basestring):
        temporal = ctx.get_date_parser().auto(value)
        if temporal is not None:
            return to_datetime(temporal, ctx)
    elif type(value) == datetime.date:
        return ctx.timezone.localize(datetime.datetime.combine(value, datetime.time(0, 0)))
    elif isinstance(value, datetime.datetime):
        return value.astimezone(ctx.timezone)

    raise EvaluationError("Can't convert '%s' to a datetime" % unicode(value))


def to_date_or_datetime(value, ctx):
    """
    Tries conversion of any value to a date or datetime
    """
    if isinstance(value, basestring):
        temporal = ctx.get_date_parser().auto(value)
        if temporal is not None:
            return temporal
    elif type(value) == datetime.date:
        return value
    elif isinstance(value, datetime.datetime):
        return value.astimezone(ctx.timezone)

    raise EvaluationError("Can't convert '%s' to a date or datetime" % unicode(value))


def to_time(value, ctx):
    """
    Tries conversion of any value to a time
    """
    if isinstance(value, basestring):
        time = ctx.get_date_parser().time(value)
        if time is not None:
            return time
    elif isinstance(value, datetime.time):
        return value
    elif isinstance(value, datetime.datetime):
        return value.astimezone(ctx.timezone).time()

    raise EvaluationError("Can't convert '%s' to a time" % unicode(value))


def to_same(value1, value2, ctx):
    """
    Converts a pair of arguments to their most-likely types. This deviates from Excel which doesn't auto convert values
    but is necessary for us to intuitively handle contact fields which don't use the correct value type
    """
    if type(value1) == type(value2):
        return value1, value2

    try:
        # try converting to two decimals
        return to_decimal(value1, ctx), to_decimal(value2, ctx)
    except EvaluationError:
        pass

    try:
        # try converting to two dates
        return to_date_or_datetime(value1, ctx), to_date_or_datetime(value2, ctx)
    except EvaluationError:
        pass

    # try converting to two strings
    return to_string(value1, ctx), to_string(value2, ctx)


def to_repr(value, ctx):
    """
    Converts a value back to its representation form, e.g. x -> "x"
    """
    as_string = to_string(value, ctx)

    if isinstance(value, basestring) or isinstance(value, datetime.date) or isinstance(value, datetime.time):
        as_string = as_string.replace('"', '""')  # escape quotes by doubling
        as_string = '"%s"' % as_string

    return as_string


def format_decimal(decimal):
    """
    Formats a decimal number using the same precision as Excel
    :param decimal: the decimal value
    :return: the formatted string value
    """
    getcontext().rounding = ROUND_HALF_UP

    # strip trailing zeros
    normalized = decimal.normalize()
    sign, digits, exponent = normalized.as_tuple()
    if exponent >= 1:
        normalized = normalized.quantize(1)
        sign, digits, exponent = normalized.as_tuple()

    int_digits = len(digits) + exponent
    fractional_digits = min(max(10 - int_digits, 0), -exponent)

    normalized = normalized.quantize(Decimal(10) ** -fractional_digits)

    return unicode(normalized)



# Generated from Excellent.g4 by ANTLR 4.5.1
# encoding: utf-8
from __future__ import print_function
from antlr4 import *
from io import StringIO

def serializedATN():
    with StringIO() as buf:
        buf.write(u"\3\u0430\ud6d1\u8206\uad2d\u4417\uaef1\u8d80\uaadd\3")
        buf.write(u"\30D\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\3\2\3\2\3\2\3\3")
        buf.write(u"\3\3\3\3\3\3\3\3\3\3\5\3\24\n\3\3\3\3\3\3\3\3\3\3\3\3")
        buf.write(u"\3\3\3\3\3\3\3\3\3\3\3\5\3!\n\3\3\3\3\3\3\3\3\3\3\3\3")
        buf.write(u"\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\3\7")
        buf.write(u"\3\65\n\3\f\3\16\38\13\3\3\4\3\4\3\5\3\5\3\5\7\5?\n\5")
        buf.write(u"\f\5\16\5B\13\5\3\5\2\3\4\6\2\4\6\b\2\7\3\2\b\t\3\2\6")
        buf.write(u"\7\3\2\r\20\3\2\13\f\3\2\24\26N\2\n\3\2\2\2\4 \3\2\2")
        buf.write(u"\2\69\3\2\2\2\b;\3\2\2\2\n\13\5\4\3\2\13\f\7\2\2\3\f")
        buf.write(u"\3\3\2\2\2\r\16\b\3\1\2\16\17\7\7\2\2\17!\5\4\3\17\20")
        buf.write(u"\21\5\6\4\2\21\23\7\4\2\2\22\24\5\b\5\2\23\22\3\2\2\2")
        buf.write(u"\23\24\3\2\2\2\24\25\3\2\2\2\25\26\7\5\2\2\26!\3\2\2")
        buf.write(u"\2\27!\7\23\2\2\30!\7\22\2\2\31!\7\24\2\2\32!\7\25\2")
        buf.write(u"\2\33!\7\26\2\2\34\35\7\4\2\2\35\36\5\4\3\2\36\37\7\5")
        buf.write(u"\2\2\37!\3\2\2\2 \r\3\2\2\2 \20\3\2\2\2 \27\3\2\2\2 ")
        buf.write(u"\30\3\2\2\2 \31\3\2\2\2 \32\3\2\2\2 \33\3\2\2\2 \34\3")
        buf.write(u"\2\2\2!\66\3\2\2\2\"#\f\16\2\2#$\7\n\2\2$\65\5\4\3\17")
        buf.write(u"%&\f\r\2\2&\'\t\2\2\2\'\65\5\4\3\16()\f\f\2\2)*\t\3\2")
        buf.write(u"\2*\65\5\4\3\r+,\f\13\2\2,-\t\4\2\2-\65\5\4\3\f./\f\n")
        buf.write(u"\2\2/\60\t\5\2\2\60\65\5\4\3\13\61\62\f\t\2\2\62\63\7")
        buf.write(u"\21\2\2\63\65\5\4\3\n\64\"\3\2\2\2\64%\3\2\2\2\64(\3")
        buf.write(u"\2\2\2\64+\3\2\2\2\64.\3\2\2\2\64\61\3\2\2\2\658\3\2")
        buf.write(u"\2\2\66\64\3\2\2\2\66\67\3\2\2\2\67\5\3\2\2\28\66\3\2")
        buf.write(u"\2\29:\t\6\2\2:\7\3\2\2\2;@\5\4\3\2<=\7\3\2\2=?\5\4\3")
        buf.write(u"\2><\3\2\2\2?B\3\2\2\2@>\3\2\2\2@A\3\2\2\2A\t\3\2\2\2")
        buf.write(u"B@\3\2\2\2\7\23 \64\66@")
        return buf.getvalue()


class ExcellentParser ( Parser ):

    grammarFileName = "Excellent.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ u"<INVALID>", u"','", u"'('", u"')'", u"'+'", u"'-'", 
                     u"'*'", u"'/'", u"'^'", u"'='", u"'<>'", u"'<='", u"'<'", 
                     u"'>='", u"'>'", u"'&'" ]

    symbolicNames = [ u"<INVALID>", u"COMMA", u"LPAREN", u"RPAREN", u"PLUS", 
                      u"MINUS", u"TIMES", u"DIVIDE", u"EXPONENT", u"EQ", 
                      u"NEQ", u"LTE", u"LT", u"GTE", u"GT", u"AMPERSAND", 
                      u"DECIMAL", u"STRING", u"TRUE", u"FALSE", u"NAME", 
                      u"WS", u"ERROR" ]

    RULE_parse = 0
    RULE_expression = 1
    RULE_fnname = 2
    RULE_parameters = 3

    ruleNames =  [ u"parse", u"expression", u"fnname", u"parameters" ]

    EOF = Token.EOF
    COMMA=1
    LPAREN=2
    RPAREN=3
    PLUS=4
    MINUS=5
    TIMES=6
    DIVIDE=7
    EXPONENT=8
    EQ=9
    NEQ=10
    LTE=11
    LT=12
    GTE=13
    GT=14
    AMPERSAND=15
    DECIMAL=16
    STRING=17
    TRUE=18
    FALSE=19
    NAME=20
    WS=21
    ERROR=22

    def __init__(self, input):
        super(ExcellentParser, self).__init__(input)
        self.checkVersion("4.5.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None



    class ParseContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(ExcellentParser.ParseContext, self).__init__(parent, invokingState)
            self.parser = parser

        def expression(self):
            return self.getTypedRuleContext(ExcellentParser.ExpressionContext,0)


        def EOF(self):
            return self.getToken(ExcellentParser.EOF, 0)

        def getRuleIndex(self):
            return ExcellentParser.RULE_parse

        def accept(self, visitor):
            if hasattr(visitor, "visitParse"):
                return visitor.visitParse(self)
            else:
                return visitor.visitChildren(self)




    def parse(self):

        localctx = ExcellentParser.ParseContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_parse)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 8
            self.expression(0)
            self.state = 9
            self.match(ExcellentParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ExpressionContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(ExcellentParser.ExpressionContext, self).__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return ExcellentParser.RULE_expression

     
        def copyFrom(self, ctx):
            super(ExcellentParser.ExpressionContext, self).copyFrom(ctx)


    class DecimalLiteralContext(ExpressionContext):

        def __init__(self, parser, ctx): # actually a ExcellentParser.ExpressionContext)
            super(ExcellentParser.DecimalLiteralContext, self).__init__(parser)
            self.copyFrom(ctx)

        def DECIMAL(self):
            return self.getToken(ExcellentParser.DECIMAL, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitDecimalLiteral"):
                return visitor.visitDecimalLiteral(self)
            else:
                return visitor.visitChildren(self)


    class ParenthesesContext(ExpressionContext):

        def __init__(self, parser, ctx): # actually a ExcellentParser.ExpressionContext)
            super(ExcellentParser.ParenthesesContext, self).__init__(parser)
            self.copyFrom(ctx)

        def LPAREN(self):
            return self.getToken(ExcellentParser.LPAREN, 0)
        def expression(self):
            return self.getTypedRuleContext(ExcellentParser.ExpressionContext,0)

        def RPAREN(self):
            return self.getToken(ExcellentParser.RPAREN, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitParentheses"):
                return visitor.visitParentheses(self)
            else:
                return visitor.visitChildren(self)


    class NegationContext(ExpressionContext):

        def __init__(self, parser, ctx): # actually a ExcellentParser.ExpressionContext)
            super(ExcellentParser.NegationContext, self).__init__(parser)
            self.copyFrom(ctx)

        def MINUS(self):
            return self.getToken(ExcellentParser.MINUS, 0)
        def expression(self):
            return self.getTypedRuleContext(ExcellentParser.ExpressionContext,0)


        def accept(self, visitor):
            if hasattr(visitor, "visitNegation"):
                return visitor.visitNegation(self)
            else:
                return visitor.visitChildren(self)


    class ExponentExpressionContext(ExpressionContext):

        def __init__(self, parser, ctx): # actually a ExcellentParser.ExpressionContext)
            super(ExcellentParser.ExponentExpressionContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expression(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(ExcellentParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(ExcellentParser.ExpressionContext,i)

        def EXPONENT(self):
            return self.getToken(ExcellentParser.EXPONENT, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitExponentExpression"):
                return visitor.visitExponentExpression(self)
            else:
                return visitor.visitChildren(self)


    class AdditionOrSubtractionExpressionContext(ExpressionContext):

        def __init__(self, parser, ctx): # actually a ExcellentParser.ExpressionContext)
            super(ExcellentParser.AdditionOrSubtractionExpressionContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expression(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(ExcellentParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(ExcellentParser.ExpressionContext,i)

        def PLUS(self):
            return self.getToken(ExcellentParser.PLUS, 0)
        def MINUS(self):
            return self.getToken(ExcellentParser.MINUS, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitAdditionOrSubtractionExpression"):
                return visitor.visitAdditionOrSubtractionExpression(self)
            else:
                return visitor.visitChildren(self)


    class FalseContext(ExpressionContext):

        def __init__(self, parser, ctx): # actually a ExcellentParser.ExpressionContext)
            super(ExcellentParser.FalseContext, self).__init__(parser)
            self.copyFrom(ctx)

        def FALSE(self):
            return self.getToken(ExcellentParser.FALSE, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitFalse"):
                return visitor.visitFalse(self)
            else:
                return visitor.visitChildren(self)


    class ContextReferenceContext(ExpressionContext):

        def __init__(self, parser, ctx): # actually a ExcellentParser.ExpressionContext)
            super(ExcellentParser.ContextReferenceContext, self).__init__(parser)
            self.copyFrom(ctx)

        def NAME(self):
            return self.getToken(ExcellentParser.NAME, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitContextReference"):
                return visitor.visitContextReference(self)
            else:
                return visitor.visitChildren(self)


    class ComparisonExpressionContext(ExpressionContext):

        def __init__(self, parser, ctx): # actually a ExcellentParser.ExpressionContext)
            super(ExcellentParser.ComparisonExpressionContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expression(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(ExcellentParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(ExcellentParser.ExpressionContext,i)

        def LTE(self):
            return self.getToken(ExcellentParser.LTE, 0)
        def LT(self):
            return self.getToken(ExcellentParser.LT, 0)
        def GTE(self):
            return self.getToken(ExcellentParser.GTE, 0)
        def GT(self):
            return self.getToken(ExcellentParser.GT, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitComparisonExpression"):
                return visitor.visitComparisonExpression(self)
            else:
                return visitor.visitChildren(self)


    class ConcatenationContext(ExpressionContext):

        def __init__(self, parser, ctx): # actually a ExcellentParser.ExpressionContext)
            super(ExcellentParser.ConcatenationContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expression(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(ExcellentParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(ExcellentParser.ExpressionContext,i)

        def AMPERSAND(self):
            return self.getToken(ExcellentParser.AMPERSAND, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitConcatenation"):
                return visitor.visitConcatenation(self)
            else:
                return visitor.visitChildren(self)


    class StringLiteralContext(ExpressionContext):

        def __init__(self, parser, ctx): # actually a ExcellentParser.ExpressionContext)
            super(ExcellentParser.StringLiteralContext, self).__init__(parser)
            self.copyFrom(ctx)

        def STRING(self):
            return self.getToken(ExcellentParser.STRING, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitStringLiteral"):
                return visitor.visitStringLiteral(self)
            else:
                return visitor.visitChildren(self)


    class FunctionCallContext(ExpressionContext):

        def __init__(self, parser, ctx): # actually a ExcellentParser.ExpressionContext)
            super(ExcellentParser.FunctionCallContext, self).__init__(parser)
            self.copyFrom(ctx)

        def fnname(self):
            return self.getTypedRuleContext(ExcellentParser.FnnameContext,0)

        def LPAREN(self):
            return self.getToken(ExcellentParser.LPAREN, 0)
        def RPAREN(self):
            return self.getToken(ExcellentParser.RPAREN, 0)
        def parameters(self):
            return self.getTypedRuleContext(ExcellentParser.ParametersContext,0)


        def accept(self, visitor):
            if hasattr(visitor, "visitFunctionCall"):
                return visitor.visitFunctionCall(self)
            else:
                return visitor.visitChildren(self)


    class TrueContext(ExpressionContext):

        def __init__(self, parser, ctx): # actually a ExcellentParser.ExpressionContext)
            super(ExcellentParser.TrueContext, self).__init__(parser)
            self.copyFrom(ctx)

        def TRUE(self):
            return self.getToken(ExcellentParser.TRUE, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitTrue"):
                return visitor.visitTrue(self)
            else:
                return visitor.visitChildren(self)


    class EqualityExpressionContext(ExpressionContext):

        def __init__(self, parser, ctx): # actually a ExcellentParser.ExpressionContext)
            super(ExcellentParser.EqualityExpressionContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expression(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(ExcellentParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(ExcellentParser.ExpressionContext,i)

        def EQ(self):
            return self.getToken(ExcellentParser.EQ, 0)
        def NEQ(self):
            return self.getToken(ExcellentParser.NEQ, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitEqualityExpression"):
                return visitor.visitEqualityExpression(self)
            else:
                return visitor.visitChildren(self)


    class MultiplicationOrDivisionExpressionContext(ExpressionContext):

        def __init__(self, parser, ctx): # actually a ExcellentParser.ExpressionContext)
            super(ExcellentParser.MultiplicationOrDivisionExpressionContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expression(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(ExcellentParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(ExcellentParser.ExpressionContext,i)

        def TIMES(self):
            return self.getToken(ExcellentParser.TIMES, 0)
        def DIVIDE(self):
            return self.getToken(ExcellentParser.DIVIDE, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitMultiplicationOrDivisionExpression"):
                return visitor.visitMultiplicationOrDivisionExpression(self)
            else:
                return visitor.visitChildren(self)



    def expression(self, _p=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = ExcellentParser.ExpressionContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 2
        self.enterRecursionRule(localctx, 2, self.RULE_expression, _p)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 30
            la_ = self._interp.adaptivePredict(self._input,1,self._ctx)
            if la_ == 1:
                localctx = ExcellentParser.NegationContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 12
                self.match(ExcellentParser.MINUS)
                self.state = 13
                self.expression(13)
                pass

            elif la_ == 2:
                localctx = ExcellentParser.FunctionCallContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 14
                self.fnname()
                self.state = 15
                self.match(ExcellentParser.LPAREN)
                self.state = 17
                _la = self._input.LA(1)
                if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ExcellentParser.LPAREN) | (1 << ExcellentParser.MINUS) | (1 << ExcellentParser.DECIMAL) | (1 << ExcellentParser.STRING) | (1 << ExcellentParser.TRUE) | (1 << ExcellentParser.FALSE) | (1 << ExcellentParser.NAME))) != 0):
                    self.state = 16
                    self.parameters()


                self.state = 19
                self.match(ExcellentParser.RPAREN)
                pass

            elif la_ == 3:
                localctx = ExcellentParser.StringLiteralContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 21
                self.match(ExcellentParser.STRING)
                pass

            elif la_ == 4:
                localctx = ExcellentParser.DecimalLiteralContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 22
                self.match(ExcellentParser.DECIMAL)
                pass

            elif la_ == 5:
                localctx = ExcellentParser.TrueContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 23
                self.match(ExcellentParser.TRUE)
                pass

            elif la_ == 6:
                localctx = ExcellentParser.FalseContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 24
                self.match(ExcellentParser.FALSE)
                pass

            elif la_ == 7:
                localctx = ExcellentParser.ContextReferenceContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 25
                self.match(ExcellentParser.NAME)
                pass

            elif la_ == 8:
                localctx = ExcellentParser.ParenthesesContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 26
                self.match(ExcellentParser.LPAREN)
                self.state = 27
                self.expression(0)
                self.state = 28
                self.match(ExcellentParser.RPAREN)
                pass


            self._ctx.stop = self._input.LT(-1)
            self.state = 52
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,3,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 50
                    la_ = self._interp.adaptivePredict(self._input,2,self._ctx)
                    if la_ == 1:
                        localctx = ExcellentParser.ExponentExpressionContext(self, ExcellentParser.ExpressionContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expression)
                        self.state = 32
                        if not self.precpred(self._ctx, 12):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 12)")
                        self.state = 33
                        self.match(ExcellentParser.EXPONENT)
                        self.state = 34
                        self.expression(13)
                        pass

                    elif la_ == 2:
                        localctx = ExcellentParser.MultiplicationOrDivisionExpressionContext(self, ExcellentParser.ExpressionContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expression)
                        self.state = 35
                        if not self.precpred(self._ctx, 11):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 11)")
                        self.state = 36
                        _la = self._input.LA(1)
                        if not(_la==ExcellentParser.TIMES or _la==ExcellentParser.DIVIDE):
                            self._errHandler.recoverInline(self)
                        else:
                            self.consume()
                        self.state = 37
                        self.expression(12)
                        pass

                    elif la_ == 3:
                        localctx = ExcellentParser.AdditionOrSubtractionExpressionContext(self, ExcellentParser.ExpressionContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expression)
                        self.state = 38
                        if not self.precpred(self._ctx, 10):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 10)")
                        self.state = 39
                        _la = self._input.LA(1)
                        if not(_la==ExcellentParser.PLUS or _la==ExcellentParser.MINUS):
                            self._errHandler.recoverInline(self)
                        else:
                            self.consume()
                        self.state = 40
                        self.expression(11)
                        pass

                    elif la_ == 4:
                        localctx = ExcellentParser.ComparisonExpressionContext(self, ExcellentParser.ExpressionContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expression)
                        self.state = 41
                        if not self.precpred(self._ctx, 9):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 9)")
                        self.state = 42
                        _la = self._input.LA(1)
                        if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ExcellentParser.LTE) | (1 << ExcellentParser.LT) | (1 << ExcellentParser.GTE) | (1 << ExcellentParser.GT))) != 0)):
                            self._errHandler.recoverInline(self)
                        else:
                            self.consume()
                        self.state = 43
                        self.expression(10)
                        pass

                    elif la_ == 5:
                        localctx = ExcellentParser.EqualityExpressionContext(self, ExcellentParser.ExpressionContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expression)
                        self.state = 44
                        if not self.precpred(self._ctx, 8):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 8)")
                        self.state = 45
                        _la = self._input.LA(1)
                        if not(_la==ExcellentParser.EQ or _la==ExcellentParser.NEQ):
                            self._errHandler.recoverInline(self)
                        else:
                            self.consume()
                        self.state = 46
                        self.expression(9)
                        pass

                    elif la_ == 6:
                        localctx = ExcellentParser.ConcatenationContext(self, ExcellentParser.ExpressionContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expression)
                        self.state = 47
                        if not self.precpred(self._ctx, 7):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 7)")
                        self.state = 48
                        self.match(ExcellentParser.AMPERSAND)
                        self.state = 49
                        self.expression(8)
                        pass

             
                self.state = 54
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,3,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class FnnameContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(ExcellentParser.FnnameContext, self).__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(ExcellentParser.NAME, 0)

        def TRUE(self):
            return self.getToken(ExcellentParser.TRUE, 0)

        def FALSE(self):
            return self.getToken(ExcellentParser.FALSE, 0)

        def getRuleIndex(self):
            return ExcellentParser.RULE_fnname

        def accept(self, visitor):
            if hasattr(visitor, "visitFnname"):
                return visitor.visitFnname(self)
            else:
                return visitor.visitChildren(self)




    def fnname(self):

        localctx = ExcellentParser.FnnameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_fnname)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 55
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ExcellentParser.TRUE) | (1 << ExcellentParser.FALSE) | (1 << ExcellentParser.NAME))) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ParametersContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(ExcellentParser.ParametersContext, self).__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return ExcellentParser.RULE_parameters

     
        def copyFrom(self, ctx):
            super(ExcellentParser.ParametersContext, self).copyFrom(ctx)



    class FunctionParametersContext(ParametersContext):

        def __init__(self, parser, ctx): # actually a ExcellentParser.ParametersContext)
            super(ExcellentParser.FunctionParametersContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expression(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(ExcellentParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(ExcellentParser.ExpressionContext,i)

        def COMMA(self, i=None):
            if i is None:
                return self.getTokens(ExcellentParser.COMMA)
            else:
                return self.getToken(ExcellentParser.COMMA, i)

        def accept(self, visitor):
            if hasattr(visitor, "visitFunctionParameters"):
                return visitor.visitFunctionParameters(self)
            else:
                return visitor.visitChildren(self)



    def parameters(self):

        localctx = ExcellentParser.ParametersContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_parameters)
        self._la = 0 # Token type
        try:
            localctx = ExcellentParser.FunctionParametersContext(self, localctx)
            self.enterOuterAlt(localctx, 1)
            self.state = 57
            self.expression(0)
            self.state = 62
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==ExcellentParser.COMMA:
                self.state = 58
                self.match(ExcellentParser.COMMA)
                self.state = 59
                self.expression(0)
                self.state = 64
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



    def sempred(self, localctx, ruleIndex, predIndex):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[1] = self.expression_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def expression_sempred(self, localctx, predIndex):
            if predIndex == 0:
                return self.precpred(self._ctx, 12)
         

            if predIndex == 1:
                return self.precpred(self._ctx, 11)
         

            if predIndex == 2:
                return self.precpred(self._ctx, 10)
         

            if predIndex == 3:
                return self.precpred(self._ctx, 9)
         

            if predIndex == 4:
                return self.precpred(self._ctx, 8)
         

            if predIndex == 5:
                return self.precpred(self._ctx, 7)
         







from setuptools import setup, find_packages


def _is_requirement(line):
    """Returns whether the line is a valid package requirement."""
    line = line.strip()
    return line and not (line.startswith("-r") or line.startswith("#"))


def _read_requirements(filename):
    """Returns a list of package requirements read from the file."""
    requirements_file = open(filename).read()
    return [line.strip() for line in requirements_file.splitlines()
            if _is_requirement(line)]


required_packages = _read_requirements("requirements/base.txt")
test_packages = _read_requirements("requirements/tests.txt")

setup(
    name='rapidpro-expressions',
    version='1.1.2',
    description='Python implementation of the RapidPro expression and templating system',
    url='https://github.com/rapidpro/flows',

    author='Nyaruka',
    author_email='code@nyaruka.com',

    license='BSD',

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2',
    ],

    keywords='rapidpro templating',
    packages=find_packages(),
    package_data={'temba_expressions': ['month.aliases']},
    install_requires=required_packages,

    test_suite='nose.collector',
    tests_require=required_packages + test_packages,
)
